#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <pthread.h>
#include "serial.h"

/*unsigned char *** images;
double *** mass;
double *** dsmass;*/

unsigned char images[NUMSOURCE][MAXCHAR][DIM2];
double mass[NUMSOURCE][MAXCHAR];
double dsmass[NUMSOURCE][MAXCHAR+1];
unsigned char myimage[DIM2];

pthread_barrier_t barrier;

int noise_thresh;

typedef struct map_data {
    int k;
} map_data;
typedef struct reduce_data {
    int a;
    int b;
} reduce_data;

inline int myrand (int min, int max) {
    return min + rand() / (RAND_MAX / (max - min) + 1);
}

int mychar, mysource;

double corr2 (unsigned char* a, unsigned char* b);

inline int nextlp2(int x) {
    if (x==0)
        return 0;

    x |= (x>>1);
    x |= (x>>2);
    x |= (x>>4);
    x |= (x>>8);
    x |= (x>>16);
    return  x - (x>>1);

}

void addNoise (unsigned char *img) {
    int i=0;
    if (noise_thresh==0)
        return;
    for (i=0; i<DIM2; i++) {
        int test=myrand(0, 100);
        if (test < noise_thresh) {
            if (rand()%2==0)
                img[i]=0;
            else
                img[i]=255;
        }
    }
}

void * mapthread( void * ptr ) {
    map_data *d = (map_data*)ptr;
    int j;
    for (j=0; j<MAXCHAR; j++) {
        double mycorr = corr2(images[d->k][j], myimage);
        mass[d->k][j] = mycorr;
    }
    DSmassC(dsmass[d->k], mass[d->k], MAXCHAR);
    pthread_barrier_wait(&barrier);
}



void * reducethread( void *ptr) {
    reduce_data *d = (reduce_data*)ptr;
   double norm =0;
   int length = MAXCHAR+1;
   double combmatrix[length][length];

   double cursum=0;
   int i, j;
   for (i=0; i<length; i++) {
        for (j=0; j<length; j++) {
            combmatrix[i][j]=dsmass[d->a][i] * dsmass[d->b][j];
        }
   }
   for (i=0; i<length; i++) {
        norm += combmatrix[i][i] + combmatrix[i][length-1] + combmatrix[length-1][i];
   }
   for (i=0; i<length-1; i++) {

        dsmass[d->a][i] = (combmatrix[i][i]+combmatrix[i][length-1]+combmatrix[(length-1)][i]) / norm;
   }
   dsmass[d->a][length-1] = (combmatrix[length-1][length-1]) / norm;
   pthread_barrier_wait(&barrier);
}



/* produces a cross correlation value a la Matlab
 *  INPUT 2 arrays of ints, assumed to be of size DIMxDIM
 *  OUTPUT cross correlation value
 */

/*   sum_m (sum_n (  ( A[m][n] - avg_a ) * ( B[m][n] - avg_b ) ) )   
a*             /   
 *   Sqrt( (sum_m(sum_n( (A[m][n]-avg_a)^2 ))) * (sum_m(sum_n( (B[m][n]-avg_b)^2 ))))
 *
*/


/* randomly shuffles an array of ints 
 * INPUT array of integers, length of array
 * OUTPUT mutates same array, elements in random order
 */
void shuffle(int *array, size_t n, size_t m)
{
    if (n > 1) 
    {
        size_t i;
        for (i = 0; i < n - 1; i++) 
        {
          size_t j = i + rand() / (RAND_MAX / (m - i) + 1);
          int t = array[j];
          array[j] = array[i];
          array[i] = t;
        }
    }
}

/* helper function to initialize array */
shufflerandoms(int * randoms, int maxsource, int numexp) {
    int i, mycount=0;
    for (i=0; i<=maxsource; i++) {
        if (i==mysource)
            mycount++;
        randoms[i]=(mycount);
        mycount++;
    }
    shuffle(randoms, maxsource, maxsource);
}





double corr2 (unsigned char *a, unsigned char *b) {
    double avg_a=0.0f, avg_b=0.0f;
    int i=0, j=0;
    
    double numerator_sum=0.0f;
    double denom_sum1=0.0f, denom_sum2=0.0f;

    double sum_a=0.0f;
    double sum_b=0.0f;

    double ret;
    for (i=0; i<DIM; i++) {
        for (j=0; j<DIM; j++) {
            sum_a += a[i*DIM+j];
            sum_b += b[i*DIM+j];
        }
    }

    avg_a = sum_a / (double)(DIM*DIM);
    avg_b = sum_b / (double)(DIM*DIM);

    for (i=0; i<DIM; i++) {
        for (j=0; j<DIM; j++) {
            numerator_sum += (a[i*DIM+j] - avg_a) * (b[i*DIM+j] - avg_b);
            denom_sum1 += (a[i*DIM+j]-avg_a) * (a[i*DIM+j]-avg_a);
            denom_sum2 += (b[i*DIM+j]-avg_b) * (b[i*DIM+j]-avg_b);
        }
    }

    ret = numerator_sum / (sqrt(denom_sum1 * denom_sum2));
    return ret;
}


int main (int argc, char ** argv) {
    // forward declarations
    char f1[256];
    char *myfile;
    struct timeval start, end;
    int i, j, k;
    int usingfile=0; // are we using an absolute file as input?
    int randoms[5420];

    mysource = -1;
    mychar = -1;
    myfile=NULL;

    for (i=1; i<argc; i++) {
        if (!strcmp(argv[i], "-s") && i+1<argc) {
            mysource = atoi(argv[++i]);

        } else if (!strcmp(argv[i], "-c") && i+1<argc) {
            mychar = atoi(argv[++i]);

        } else if (!strcmp(argv[i], "-n") && i+1<argc) {
            noise_thresh = atoi(argv[++i]);

        } else if (!strcmp(argv[i], "--myfile") && i+1<argc) {
            mysource =-1;
            mychar =-1;
            usingfile=1;
            myfile = argv[++i];
        } else {
            printf("Invalid flag.\n");
            return 1;
        }
    }

    if (myfile==NULL) {
        // it will have been assigned to argv if --myfile used...
        // thus it will only be null if --myfile is not used
        myfile = (char*)malloc(256*sizeof(char));
    }

    srand((unsigned int)time(NULL));
    shufflerandoms(randoms, 5420, NUMSOURCE);
    

    FILE *fp1;

    pthread_t threads[NUMSOURCE-1];
   
    map_data tdata[NUMSOURCE-1];
    reduce_data rdata[(NUMSOURCE+1)/2];

    int sourceskipper=0;
    
    if (!usingfile) {
        sprintf(myfile, "./font-dump/img-%d-%d.dat", mychar, mysource);
    }
    fp1 = fopen(myfile, "rb");
    if (!fp1) {
        perror("error");
        printf("%s\n", myfile);
        exit(1);
    }
    fread(myimage, sizeof(unsigned char), DIM*DIM, fp1);
    fclose(fp1);

    for (i=0; i<NUMSOURCE; i++) {
        for (j=0; j<MAXCHAR; j++) {
            if (noise_thresh==0)
                sprintf(f1, "./font-dump/img-%d-%d.dat", j, randoms[sourceskipper]);
            else
                sprintf(f1, "./font-dump-noise-%d/img-%d-%d.dat", noise_thresh, j, randoms[sourceskipper]);
            fp1 = fopen(f1, "rb");
            if (!fp1) {
                perror("error");
                printf("%s\n", f1);
                continue;
            }
            fread(images[i][j], sizeof(unsigned char), DIM*DIM, fp1);
            fclose(fp1);
        }
        sourceskipper++;
    }

    // all in memory...

    gettimeofday(&start, 0);

    pthread_barrier_init(&barrier, NULL, NUMSOURCE+1);

    for (k=0; k<NUMSOURCE; k++) {
        //launch new thread...
        tdata[k].k = k;
        pthread_create(&threads[k], NULL, (void*) &mapthread, (void *) &tdata[k]);
    }
    pthread_barrier_wait(&barrier);
    
    pthread_barrier_destroy(&barrier);

    for (k=1; k<NUMSOURCE; k++) {
        DCRsingletons(dsmass[0], dsmass[k], MAXCHAR+1);
    }

    gettimeofday(&end, 0);
    long elapsed = (end.tv_sec-start.tv_sec)*1000000+(end.tv_usec-start.tv_usec);
    printf("Time elapsed: %ld\n", elapsed);


    for(i=0; i<MAXCHAR+1; i++) {
        printf("%d\t%f\n", i, dsmass[0][i]);
    }

    return 0;

}
